/* Copyright (C) 2005-2010, Unigine Corp. All rights reserved.
 *
 * File:    sanctuary.cpp
 * Desc:    Sanctuary demo
 * Version: 1.22
 * Authors: Alexander Zaprjagaev <frustum@unigine.com>
 *          Valia Vaneeva <fattie@unigine.com>
 *
 * This file is part of the Unigine engine (http://unigine.com/).
 *
 * Your use and or redistribution of this software in source and / or
 * binary form, with or without modification, is subject to: (i) your
 * ongoing acceptance of and compliance with the terms and conditions of
 * the Unigine License Agreement; and (ii) your inclusion of this notice
 * in any version of this software that you use or redistribute.
 * A copy of the Unigine License Agreement is available by contacting
 * Unigine Corp. at http://unigine.com/
 */

#include <core/unigine.h>
#include <core/scripts/benchmark.h>
#include <core/scripts/phoronix.h>
#include <core/scripts/camera.h>
#include <core/scripts/spline.h>
#include <core/scripts/noise.h>

//#define BENCHMARK
//#define PHORONIX
//#define EDITOR

#define KEY_PAUSE			' '
#define KEY_NEXT_PATH		APP_KEY_RETURN
#define KEY_NEXT_PRESET		APP_KEY_RIGHT
#define KEY_PREV_PRESET		APP_KEY_LEFT
#define KEY_FREE_CAMERA		APP_KEY_F8
#define KEY_BENCHMARK		APP_KEY_F9

/******************************************************************************\
*
* Class declarations
*
\******************************************************************************/

/*
 */
class Preset {
	
	Preset(void settings[]);
	
	string getName();
	
	void setPreset();
	
	string name;					// preset name
	
	int brightness;					// brightness
	int contrast;					// contrast
	
	string color_name;				// color lut
	
	string material;				// player material
	
	float parameters[0];			// material parameters
};

/*
 */
class Camera {
	
	Camera(Sanctuary sanctuary);
	
	void setMode(int mode);
	int getMode();
	
	void setPause(int pause);
	int getPause();
	
	void setPath(int path);
	int getPath();
	
	void prevPath();
	void nextPath();
	
	void update();
	
	enum {
		NUM_PATH = 9,
	};
	
	Sanctuary sanctuary;
	
	int mode;								// camera mode
	int pause;								// pause toggle
	float frame;							// current frame
	float clamp = 20.0;						// clamp frames
	float fade = 30.0;						// fade frames
	float fps = 30.0;						// path fps
	
	PlayerDummy dummy;						// dummy player
	PlayerSpectator spectator;				// spectator player
	
	int path;								// current path
	Unigine::CameraPath paths[NUM_PATH];	// camera paths
	
	float dof = 1000.0;						// depth of field distance
	float adaptation = 0.25;				// depth of field adaptation time
};

/*
 */
class Flame {
	
	Flame(int num);
	
	void save();
	void restore();
	
	void setEnabled(int enable);
	int isEnabled();
	
	void update();
	
	LightOmni light;				// light
	int light_enabled;				// enabled flag
	mat4 light_transform;			// transformation
	vec4 light_color;				// color

	ObjectVolumeSphere volume;		// volume
	mat4 volume_transform;			// transformation
	
	ObjectParticles particles;		// particles
	mat4 particles_transform;		// transformation
	
	Unigine::Noise noise;			// noise generator
};

/*
 */
class Candle {
	
	Candle(int num);
	
	void save();
	void restore();
	
	void setEnabled(int enable);
	int isEnabled();
	
	void update();
	
	LightProj light;				// light
	int light_enabled;				// enabled
	mat4 light_transform;			// transformation
	vec4 light_color;				// color
	
	ObjectVolumeSphere volume;		// volume
	mat4 volume_transform;			// transformation
	
	ObjectParticles particles;		// particles
	mat4 particles_transform;		// transformation
	
	Unigine::Noise noise;			// noise generator
};

/*
 */
class Weapon {
	
	Weapon(Sanctuary sanctuary);
	
	void setEnabled(int enable);
	int isEnabled();
	
	void update();
	
	Sanctuary sanctuary;
	
	ObjectMesh candle;				// candle
	ObjectParticles particles;		// particles
	mat4 particles_transform;		// transformation
};

/*
 */
class Fireball {
	
	Fireball(int num,Sanctuary sanctuary);
	
	void save();
	void restore();
	
	void setEnabled(int enable);
	int isEnabled();
	
	void setPause(int pause);
	int getPause();
	
	void setPath(int path);
	int getPath();
	
	void prevPath();
	void nextPath();
	
	void update();
	
	enum {
		NUM_PATH = 12,
	};
	
	Sanctuary sanctuary;
	
	int pause;								// pause toggle
	float frame;							// current frame
	float fps = 18.0;						// path fps
	
	int path;								// current path
	Unigine::SplinePath paths[NUM_PATH];	// spline paths
	
	ObjectVolumeSphere volume_body;			// volume
	int volume_body_enabled;				// enabled
	mat4 volume_body_transform;				// transformation
	
	LightOmni light;						// light
	vec3 light_radius;
	float light_multiplier;
	mat4 light_transform;
	
	ObjectParticles particles;				// particles
	int particles_spawn_rate;
	mat4 particles_transform;
	float particles_attractor;
};

/*
 */
class Sun {
	
	Sun();
	
	void save();
	void restore();
	
	void setEnabled(int enable);
	int isEnabled();
	
	void reset();
	void update();
	
	LightWorld light;				// light
	int light_enabled;				// enabled
	mat4 transform;					// transformation
	float time;						// time
};

/*
 */
class Interface {
	
	Interface(Sanctuary sanctuary);
	~Interface();
	
	void init();
	void shutdown();
	
	void __save__();
	void __restore__();
	
	void settings_clicked();
	void camera_clicked();
	void fireballs_clicked();
	void postprocess_clicked();
	void preset_clicked(int preset);
	void camera_mode_clicked(int mode);
	void sound_clicked();
	void benchmark_clicked();
	void editor_clicked();
	void quit_clicked();
	
	void update();
	
	Sanctuary sanctuary;
	
	string mode_names[0];				// camera mode names
	
	WidgetVBox preset_vbox;				// preset panel
	WidgetButton preset_b[0];			// preset buttons
	WidgetVBox camera_mode_vbox;		// camera mode panel
	WidgetButton camera_mode_b[0];		// camera mode buttons
};

/*
 */
class Sanctuary {
	
	Sanctuary();
	~Sanctuary();
	
	void __save__();
	void __restore__();
	
	void setMode(int mode);
	int getMode();
	
	void setFlame(int enable);
	int getFlame();
	
	void setCandle(int enable);
	int getCandle();
	
	void setWeapon(int enable);
	int getWeapon();
	
	void setFireball(int enable);
	int getFireball();
	
	void setSun(int enable);
	int getSun();
	
	void setPreset(int preset);
	int getPreset();
	
	void prevPreset();
	void nextPreset();
	
	void runBenchmark();
	void stopBenchmark();
	int getBenchmark();
	
	void runPhoronix();
	void getPhoronix();
	
	void playSound();
	void stopSound();
	
	void update();
	
	int preset;						// preset
	Camera camera;					// camera
	
	Flame flames[2];				// flames
	Candle candles[2];				// candles
	Fireball fireballs[2];			// fireballs
	Sun sun;						// sun
	Weapon weapon;					// weapon
	
	AmbientSource sound;			// sound
	
	Interface interface;			// interface
	
	Unigine::Benchmark benchmark;	// benchmark
	Unigine::Phoronix phoronix;		// phoronix	benchmark
};

/******************************************************************************\
*
* Preset
*
\******************************************************************************/

/*
 */
Preset presets[0];

/*
 */
void Preset::__Preset__(void settings[]) {
	
	name = settings[0];
	
	brightness = settings[1];
	contrast = settings[2];
	
	color_name = settings[3];
	
	material = settings[4];
	
	for(int i = 5; i < settings.size(); i++) {
		parameters.append(settings[i]);
	}
}

/*
 */
string Preset::getName() {
	return name;
}

/*
 */
void Preset::setPreset() {
	
	// color correction
	engine.render.setColorBrightness(brightness / 127.0);
	engine.render.setColorContrast(contrast / 127.0);
	
	// color lut texture
	engine.render.setColorTextureName(color_name);
	
	// player materials
	Player player = engine.game.getPlayer();
	if(player != NULL) player.setMaterials(material);
	
	// material parameters
	string names[0];
	strsplit(material,",",names);
	foreach(string name, parameter = 0; names) {
		Material material = engine.materials.findMaterial(name);
		for(int i = 0; i < material.getNumParameters() && parameter < parameters.size(); i++, parameter++) {
			material.setParameter(i,vec4(parameters[parameter]));
		}
	}
}

/*
 */
#define DECLARE_PRESET(SETTINGS) presets.append(new Preset(SETTINGS));

DECLARE_PRESET(("default",								0, 		0,		NULL, 								"" ))
DECLARE_PRESET(("blue tone",							0, 		0,		"sanctuary/colors/blue.dds",		"" ))
DECLARE_PRESET(("orange",								0, 		0,		"sanctuary/colors/orange.dds",		"" ))
DECLARE_PRESET(("grayscale",							0, 		0,		"sanctuary/colors/grayscale.dds",	"" ))
DECLARE_PRESET(("sepia",								0, 		0,		"sanctuary/colors/sepia.dds",		"" ))
DECLARE_PRESET(("2d blur",								0, 		0,		NULL,								"post_hblur_2d,post_vblur_2d"))
DECLARE_PRESET(("radial blur",							0, 		0,		NULL,								"post_blur_radial",	0.5 ))
DECLARE_PRESET(("blue tone\nradial blur",				0, 		0,		"sanctuary/colors/blue.dds",		"post_blur_radial",	0.5 ))
DECLARE_PRESET(("sobel filter",							0, 		0,		NULL,								"post_filter_sobel",	0.3, 1.0 ))
DECLARE_PRESET(("blue tone\nsobel filter",				-67, 	-114,	"sanctuary/colors/blue.dds",		"post_filter_sobel",	0.3, 0.3 ))
DECLARE_PRESET(("sepia\nsobel filter\nradial blur",		-67, 	-114,	"sanctuary/colors/sepia.dds",		"post_filter_sobel,post_blur_radial",	0.3, 0.3, 0.5 ))

#undef DECLARE_PRESET

/******************************************************************************\
*
* Camera
*
\******************************************************************************/

/*
 */
void Camera::__Camera__(Sanctuary s) {
	
	sanctuary = s;
	
	// dummy player
	dummy = new PlayerDummy();
	dummy.setMaterials("sanctuary_post");
	engine.game.setPlayer(dummy);
	
	// spectator player
	spectator = new PlayerSpectator();
	spectator.setMaterials("sanctuary_post");
	spectator.setCollisionRadius(2.0);
	spectator.setMinVelocity(10.0);
	spectator.setMaxVelocity(30.0);
	
	// load paths
	forloop(int i = 0; NUM_PATH) {
		paths[i] = new Unigine::CameraPath(format("sanctuary/cameras/camera_%d.txt",i),1,0.01);
	}
}

/*
 */
void Camera::setMode(int m) {
	
	if(mode == m) return;
	
	// switch from guided
	if(mode == 0) {
		spectator.setFov(dummy.getFov());
		spectator.setPosition(dummy.getPosition());
		spectator.setDirection(dummy.getDirection());
		spectator.setMaterials(dummy.getMaterials());
		engine.game.setPlayer(spectator);
	}
	// switch to guided
	else if(m == 0) {
		dummy.setMaterials(spectator.getMaterials());
		engine.game.setPlayer(dummy);
	}
	
	mode = m;
}

int Camera::getMode() {
	return mode;
}

/*
 */
void Camera::setPause(int p) {
	pause = p;
}

int Camera::getPause() {
	return pause;
}

/*
 */
void Camera::setPath(int p) {
	path = ::clamp(p,0,NUM_PATH - 1);
	pause = 0;
	frame = 0.0;
}

int Camera::getPath() {
	return path;
}

void Camera::prevPath() {
	if(path-- == 0) path = NUM_PATH - 1;
	setPath(path);
}

void Camera::nextPath() {
	if(++path == NUM_PATH) path = 0;
	setPath(path);
}

/*
 */
void Camera::update() {
	
	// free mode
	if(mode == 0 && engine.editor.isLoaded() == 0) {
		
		// update frame
		if(pause == 0) frame += engine.game.getIFps() * fps;
		
		// switch to next path
		Unigine::CameraPath p = paths[path];
		if(frame > p.getNumFrames() - clamp * 2.0) {
			nextPath();
			p = paths[path];
		}
		
		// update path
		p.setFrame(dummy,frame + clamp);
		
		// fade
		float f = frame / fade;
		if(f > 1.0) f = (p.getNumFrames() - clamp * 2.0 - frame) / fade;
		if(f > 1.0) f = 1.0;
		engine.render.setFadeColor(f - 1.0);
	}
	
	// scripted mode
	else {
		engine.render.setFadeColor(0.0);
	}
	
	// dynamic depth of field
	int ret[4];
	Player p = engine.editor.getPlayer();
	if(p == NULL) p = engine.game.getPlayer();
	if(engine.world.getIntersection(p.getPosition(),p.getPosition() + p.getDirection() * 1000.0,~0,ret) != NULL) {
		dof = lerp(dof,length(ret[0] - p.getPosition()),saturate(engine.game.getIFps() / adaptation));
	}
	engine.render.setDOFFocalDistance(dof);
}

/******************************************************************************\
*
* Flame
*
\******************************************************************************/

/*
 */
void Flame::__Flame__(int num) {
	
	noise = new Unigine::Noise(2.0);
	
	light = node_editor("light_flame_" + num);
	volume = node_editor("volume_flame_" + num);
	particles = node_editor("particles_flame_" + num);
	
	save();
}

/*
 */
void Flame::save() {
	
	light_enabled = light.isEnabled();
	light_transform = light.getWorldTransform();
	light_color = light.getColor();
	
	volume_transform = volume.getWorldTransform();
	
	if(particles.getNumForces() == 1) {
		particles_transform = particles.getForceTransform(0);
	}
}

void Flame::restore() {
	
	light.setEnabled(light_enabled);
	light.setWorldTransform(light_transform);
	light.setColor(light_color);
	
	volume.setWorldTransform(volume_transform);
	
	if(particles.getNumForces() == 1) {
		particles.setForceTransform(0,particles_transform);
	}
}

/*
 */
void Flame::setEnabled(int enable) {
	light.setEnabled(enable);
}

int Flame::isEnabled() {
	return light.isEnabled();
}

/*
 */
void Flame::update() {
	
	vec3 n = noise.get3();
	
	light.setWorldTransform(light_transform * translate(n * 0.6));
	light.setColor(light_color * (1.0 + n.z * 0.6));
	
	volume.setWorldTransform(volume_transform * translate(n * 2.0));
	
	if(particles.getNumForces() == 1) {
		particles.setForceTransform(0,particles_transform * translate(n * 8.0));
	}
}

/******************************************************************************\
*
* Candle
*
\******************************************************************************/

/*
 */
void Candle::__Candle__(int num) {
	
	noise = new Unigine::Noise(2.0);
	
	light = node_editor("light_candle_" + num);
	volume = node_editor("volume_candle_" + num);
	particles = node_editor("particles_candle_" + num);
	
	save();
}

/*
 */
void Candle::save() {
	
	light_enabled = light.isEnabled();
	light_transform = light.getWorldTransform();
	light_color = light.getColor();
	
	volume_transform = volume.getWorldTransform();
	
	if(particles.getNumForces() == 1) {
		particles_transform = particles.getForceTransform(0);
	}
}

void Candle::restore() {
	
	light.setEnabled(light_enabled);
	light.setWorldTransform(light_transform);
	light.setColor(light_color);
	
	volume.setWorldTransform(volume_transform);
	
	if(particles.getNumForces() == 1) {
		particles.setForceTransform(0,particles_transform);
	}
}

/*
 */
void Candle::setEnabled(int enable) {
	light.setEnabled(enable);
}

int Candle::isEnabled() {
	return light.isEnabled();
}

/*
 */
void Candle::update() {
	
	vec3 n = noise.get3();
	
	light.setWorldTransform(light_transform * translate(n * 0.2));
	light.setColor(light_color * (1.0 + n.z * 0.4));
	
	volume.setWorldTransform(volume_transform * translate(n * 0.4));
	
	if(particles.getNumForces() == 1) {
		particles.setForceTransform(0,particles_transform * translate(n * 2.0));
	}
}

/******************************************************************************\
*
* Weapon
*
\******************************************************************************/

/*
 */
void Weapon::__Weapon__(Sanctuary s) {
	
	sanctuary = s;
	
	candle = node_editor("weapon_candle");
	particles = node_editor("weapon_particles");
	particles_transform = particles.getTransform();
}

/*
 */
void Weapon::setEnabled(int enable) {
	candle.setEnabled(enable);
	particles.setEnabled(enable);
}

int Weapon::isEnabled() {
	return candle.isEnabled();
}

/*
 */
void Weapon::update() {
	
	// turn off the lights
	sanctuary.setFlame(0);
	sanctuary.setSun(0);
	
	mat4 old_transform = candle.getWorldTransform();
	
	// target transformation for interpolation
	Player player = engine.game.getPlayer();
	mat4 offset = translate(3.5,-10.0,-10.0) * rotateY(-150.0) * rotateX(-80.0) * rotateZ(50.0);
	mat4 transform = player.getIModelview() * offset;
	
	// new position
	vec3 translation = lerp(old_transform.m03m13m23,transform.m03m13m23,engine.game.getIFps() * 8.0);
	
	// don't let the weapon fall behind
	translation = clamp(translation,transform.m03m13m23 - vec3(4.0,4.0,4.0),transform.m03m13m23 + vec3(4.0,4.0,4.0));

	transform.m03m13m23 = translation;
	
	candle.setWorldTransform(transform);
	particles.setPreserveWorldTransform(transform * particles_transform);
}

/******************************************************************************\
*
* Sun
*
\******************************************************************************/

/*
 */
void Sun::__Sun__() {
	
	light = node_editor("light_sun");
	time = 0.0f;
	
	save();
}

/*
 */
void Sun::save() {
	
	light_enabled = light.isEnabled();
	transform = light.getWorldTransform();
}

void Sun::restore() {
	
	light.setEnabled(light_enabled);
	light.setWorldTransform(transform);
}

/*
 */
void Sun::setEnabled(int enable) {
	light.setEnabled(enable);
}

int Sun::isEnabled() {
	return light.isEnabled();
}

/*
 */
void Sun::reset() {
	time = 0.0f;
}

/*
 */
void Sun::update() {
	time += engine.game.getIFps();
	float angle = 360.0 * time / 60.0 / 20.0;
	light.setWorldTransform(rotate(vec3(0.0,0.0,1.0),angle) * transform);
}

/******************************************************************************\
*
* Fireball
*
\******************************************************************************/

/*
 */
void Fireball::__Fireball__(int num,Sanctuary s) {
	
	sanctuary = s;
	
	// first path
	path = int(NUM_PATH / 2) * num;
	
	// load paths
	forloop(int i = 0; NUM_PATH) {
		paths[i] = new Unigine::SplinePath(format("sanctuary/splines/fireball_%02d.txt",i),0,0.01,vec3(0,0,1));
	}
	
	volume_body = node_editor("volume_fireball_body_" + num);
	
	light = node_editor("light_fireball_" + num);
	
	particles = node_editor("flame_fireball_" + num);
	
	save();
}

void Fireball::save() {
	
	volume_body_enabled = volume_body.isEnabled();
	volume_body_transform = volume_body.getWorldTransform();
	
	particles_spawn_rate = particles.getSpawnRate();
	particles_transform = particles.getForceTransform(2);
	particles_attractor = particles.getForceAttractor(2);
	
	light_radius = light.getRadius();
	light_multiplier = light.getMultiplier();
	light_transform = light.getWorldTransform();
}

void Fireball::restore() {
	
	volume_body.setEnabled(volume_body_enabled);
	volume_body.setWorldTransform(volume_body_transform);
	
	particles.setSpawnRate(particles_spawn_rate);
	particles.setForceTransform(2,particles_transform);
	particles.setForceAttractor(2,particles_attractor);
	
	light.setRadius(light_radius);
	light.setMultiplier(light_multiplier);
	light.setWorldTransform(light_transform);
}

/*
 */
void Fireball::setEnabled(int enable) {
	volume_body.setEnabled(enable);
}

int Fireball::isEnabled() {
	return volume_body.isEnabled();
}

/*
 */
void Fireball::setPause(int p) {
	pause = p;
}

int Fireball::getPause() {
	return pause;
}

/*
 */
void Fireball::setPath(int p) {
	path = ::clamp(p,0,NUM_PATH - 1);
	pause = 0;
	frame = 0.0;
}

int Fireball::getPath() {
	return path;
}

void Fireball::prevPath() {
	if(path-- == 0) path = NUM_PATH - 1;
	setPath(path);
}

void Fireball::nextPath() {
	if(++path == NUM_PATH) path = 0;
	setPath(path);
}

/*
 */
void Fireball::update() {
	
	// fireball is disabled
	if(volume_body.isEnabled() == 0) {
		
		// enable lights
		sanctuary.setFlame(1);
		sanctuary.setCandle(1);
		sanctuary.setSun(1);
	}
	
	// fireball is enabled
	else {
		
		// disable lights
		sanctuary.setFlame(0);
		sanctuary.setCandle(1);
		sanctuary.setSun(0);
		
		// update frame
		if(pause == 0) frame += engine.game.getIFps() * fps;
		
		// switch to next path
		Unigine::SplinePath p = paths[path];
		if(frame > p.getNumFrames()) {
			nextPath();
			p = paths[path];
		}
		
		// update path
		p.setFrame(volume_body,frame);
		
		// force update particles
		engine.world.addUpdateNode(particles);
		
		// particles position
		vec3 position = particles.getWorldTransform() * vec3_zero;
		
		// bowls positions
		vec3 positions[] = ( vec3(25.5,2.6,24.0), vec3(25.5,-113.0,24.0), vec3(129.8,-113.0,24.0), vec3(129.8,2.6,24.0) );
		
		float morph = 0.0;
		foreach(vec3 p; positions) morph += saturate(1.0 - length2(position - p) / 120.0);
		morph = saturate(morph);
		
		// morph particles parameters
		particles.setSpawnRate(lerp(400,200,morph));
		particles.setForceTransform(2,translate(position + vec3(0,0,20)));
		particles.setForceAttractor(2,lerp(0,-400.0,morph));
		
		// morph lights parameters
		light.setRadius(lerp(vec3(90),vec3(110),morph));
		light.setMultiplier(lerp(1.7,1.9,morph));
		light.setWorldTransform(translate(position + vec3(0.0,0.0,lerp(0.0,5.0,morph))));
	}
}

/******************************************************************************\
*
* Interface
*
\******************************************************************************/

/*
 */
Gui gui;
WidgetVBox sanctuary_vbox;
WidgetVBox postprocess_vbox;
WidgetVBox camera_vbox;
WidgetButton settings_b;
WidgetButton camera_b;
WidgetButton fireballs_b;
WidgetButton postprocess_b;
WidgetButton benchmark_b;
WidgetButton editor_b;
WidgetButton sound_b;
Interface interface;

/*
 */
void Interface::__Interface__(Sanctuary s) {
	
	sanctuary = s;
	
	// singleton
	assert(interface == NULL);
	interface = this;
	
	// fill button names for camera modes
	mode_names = ( "Guided", "Free", "Candle" );
	
	init();
}

void Interface::__~Interface__() {
	
	shutdown();
}

/*
 */
void Interface::init() {
	
	// gui
	gui = engine.getGui();
	
	// user interface
	engine.gui.addUserInterface(gui,"sanctuary/sanctuary.ui");
	
	// editor button
	#ifndef EDITOR
		editor_b.setHidden(1);
	#endif
	
	// panel
	gui.addChild(sanctuary_vbox,GUI_ALIGN_OVERLAP);
	
	// preset panel
	preset_vbox = new WidgetVBox(gui);
	
	// preset buttons
	preset_b.clear();
	foreach(Preset preset, i = 0; presets; i++) {
		WidgetButton button = new WidgetButton(gui,preset.getName());
		button.setTextAlign(GUI_ALIGN_CENTER);
		preset_vbox.addChild(button,GUI_ALIGN_EXPAND);
		preset_b.append(button);
		button.setCallback(GUI_CLICKED,"preset_clicked",i);
	}
	
	// camera mode panel
	camera_mode_vbox = new WidgetVBox(gui);
	
	// camera mode buttons
	camera_mode_b.clear();
	for(int i = 0; i < 3; i++) {
		WidgetButton button = new WidgetButton(gui,mode_names[i]);
		button.setTextAlign(GUI_ALIGN_CENTER);
		camera_mode_vbox.addChild(button,GUI_ALIGN_EXPAND);
		camera_mode_b.append(button);
		button.setCallback(GUI_CLICKED,"camera_mode_clicked",i);
	}
}

void Interface::shutdown() {
	
}

/*
 */ 
void Interface::__save__() {
	
}

void Interface::__restore__() {
	init();
}

/*
 */
void Interface::settings_clicked() {
	settings_b.removeFocus();
	if(settings_b.isToggled()) engine.system.call("systemShow");
	else engine.system.call("systemHide");
}

void Interface::camera_clicked() {
	camera_b.removeFocus();
	if(camera_b.isToggled()) camera_vbox.addChild(camera_mode_vbox,GUI_ALIGN_EXPAND);
	else camera_vbox.removeChild(camera_mode_vbox);
}

void Interface::fireballs_clicked() {
	fireballs_b.removeFocus();
	sanctuary.setFireball(fireballs_b.isToggled());
}

void Interface::postprocess_clicked() {
	postprocess_b.removeFocus();
	if(postprocess_b.isToggled()) postprocess_vbox.addChild(preset_vbox,GUI_ALIGN_EXPAND);
	else postprocess_vbox.removeChild(preset_vbox);
}

void Interface::preset_clicked(int preset) {
	postprocess_b.setToggled(0);
	sanctuary.setPreset(preset);
}

void Interface::camera_mode_clicked(int mode) {
	camera_b.setToggled(0);
	sanctuary.setMode(mode);
}

void Interface::sound_clicked() {
	sound_b.removeFocus();
	if(sound_b.isToggled()) sanctuary.playSound();
	else sanctuary.stopSound();
}

void Interface::benchmark_clicked() {
	benchmark_b.removeFocus();
	if(benchmark_b.isToggled()) sanctuary.runBenchmark();
	else sanctuary.stopBenchmark();
}

void Interface::editor_clicked() {
	editor_b.removeFocus();
	if(editor_b.isToggled()) engine.console.run("editor_load");
	else engine.console.run("editor_quit");
}

void Interface::quit_clicked() {
	engine.console.run("quit");
}

/*
 */
void Interface::update() {
	
	// update settings button
	if(settings_b.isToggled() != engine.system.call("systemGetActivity")) {
		if(settings_b.isFocused() == 0) {
			settings_b.setCallbackEnabled(GUI_CLICKED,0);
			settings_b.setToggled(engine.system.call("systemGetActivity"));
			settings_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// update camera button
	if(camera_b.isToggled()) {
		int mode = sanctuary.getMode();
		foreach(WidgetButton button, i = 0; camera_mode_b; i++) {
			button.setFontColor((mode == i) ? vec4_one : vec4(0.6));
		}
	}
	
	// update fireball button
	if(fireballs_b.isToggled() != sanctuary.getFireball()) {
		if(fireballs_b.isFocused() == 0) {
			fireballs_b.setCallbackEnabled(GUI_CLICKED,0);
			fireballs_b.setToggled(sanctuary.getFireball());
			fireballs_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// update postprocess buttons
	if(postprocess_b.isToggled()) {
		int preset = sanctuary.getPreset();
		foreach(WidgetButton button, i = 0; preset_b; i++) {
			button.setFontColor((preset == i) ? vec4_one : vec4(0.6));
		}
	}
	
	// update benchmark button
	if(benchmark_b != NULL && benchmark_b.isToggled() != sanctuary.getBenchmark()) {
		if(benchmark_b.isFocused() == 0) {
			benchmark_b.setCallbackEnabled(GUI_CLICKED,0);
			benchmark_b.setToggled(sanctuary.getBenchmark());
			benchmark_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// update editor button
	if(editor_b != NULL && editor_b.isToggled() != engine.editor.isLoaded()) {
		if(editor_b.isFocused() == 0) {
			editor_b.setCallbackEnabled(GUI_CLICKED,0);
			editor_b.setToggled(engine.editor.isLoaded());
			editor_b.setCallbackEnabled(GUI_CLICKED,1);
		}
	}
	
	// demo is running
	if(engine.editor.isLoaded() == 0) {
		sanctuary_vbox.setPosition(0,0);
		benchmark_b.setEnabled(1);
	}
	// editor is loaded
	else {
		sanctuary_vbox.setPosition(0,42);
		benchmark_b.setEnabled(0);
	}
}

/*
 */
void settings_clicked() {
	interface.settings_clicked();
}

void camera_clicked() {
	interface.camera_clicked();
}

void fireballs_clicked() {
	interface.fireballs_clicked();
}

void postprocess_clicked() {
	interface.postprocess_clicked();
}

void preset_clicked(int preset) {
	interface.preset_clicked(preset);
}

void camera_mode_clicked(int mode) {
	interface.camera_mode_clicked(mode);
}

void benchmark_clicked() {
	interface.benchmark_clicked();
}

void editor_clicked() {
	interface.editor_clicked();
}

void sound_clicked() {
	interface.sound_clicked();
}

void quit_clicked() {
	interface.quit_clicked();
}

/******************************************************************************\
*
* Sanctuary
*
\******************************************************************************/

/*
 */
void Sanctuary::__Sanctuary__() {
	
	// create camera
	camera = new Camera(this);
	
	// create flames
	forloop(int i = 0; 2) flames[i] = new Flame(i);
	
	// create candles
	forloop(int i = 0; 2) candles[i] = new Candle(i);
	
	// create fireballs
	forloop(int i = 0; 2) fireballs[i] = new Fireball(i,this);
	
	// create sun
	sun = new Sun();
	
	// create weapon
	weapon = new Weapon(this);
	
	// create sound
	sound = new AmbientSource("sanctuary/sanctuary.ogg",1);
	
	// create interface
	interface = new Interface(this);
	
	// create benchmarks
	int flags = Unigine::Benchmark::RENDER_SHADERS |
		Unigine::Benchmark::RENDER_TEXTURES |
		Unigine::Benchmark::RENDER_FILTER |
		Unigine::Benchmark::RENDER_ANISOTROPY |
		Unigine::Benchmark::RENDER_TRANSLUCENT |
		Unigine::Benchmark::RENDER_PARALLAX |
		Unigine::Benchmark::RENDER_REFLECTION |
		Unigine::Benchmark::RENDER_DOF |
		Unigine::Benchmark::RENDER_HDR;
	
	benchmark = new Unigine::Benchmark("Sanctuary demo<br/>benchmark edition v2.3",42.4,flags);
	phoronix = new Unigine::Phoronix();
	
	// set preset
	if(engine.config.isExist("sanctuary_preset")) setPreset(engine.config.getInt("sanctuary_preset"));
	else setPreset(0);
	
	// play sound
	if(engine.config.isExist("sanctuary_sound")) sound_b.setToggled(engine.config.getBool("sanctuary_sound"));
	else sound_b.setToggled(1);
}

void Sanctuary::__~Sanctuary__() {
	
	engine.config.setInt("sanctuary_preset",preset);
	engine.config.setBool("sanctuary_sound",sound_b.isToggled());
}

/*
 */
void Sanctuary::__save__() {
	
}

void Sanctuary::__restore__() {
	setPreset(preset);
}

/*
 */
void Sanctuary::setMode(int mode) {
	camera.setMode(mode);
}

int Sanctuary::getMode() {
	return camera.getMode();
}

/*
 */
void Sanctuary::setFlame(int enable) {
	foreach(Flame flame; flames) flame.setEnabled(enable);
}

int Sanctuary::getFlame() {
	return flames[0].isEnabled();
}

/*
 */
void Sanctuary::setCandle(int enable) {
	foreach(Candle candle; candles) candle.setEnabled(enable);
}

int Sanctuary::getCandle() {
	return candles[0].isEnabled();
}

/*
 */
void Sanctuary::setWeapon(int enable) {
	weapon.setEnabled(enable);
}

int Sanctuary::getWeapon() {
	return weapon.isEnabled();
}


/*
 */
void Sanctuary::setFireball(int enable) {
	foreach(Fireball fireball; fireballs) fireball.setEnabled(enable);
}

int Sanctuary::getFireball() {
	return fireballs[0].isEnabled();
}

/*
 */
void Sanctuary::setSun(int enable) {
	sun.setEnabled(enable);
}

int Sanctuary::getSun() {
	return sun.isEnabled();
}

/*
 */
void Sanctuary::setPreset(int p) {
	preset = clamp(p,0,presets.size() - 1);
	presets[preset].setPreset();
}

int Sanctuary::getPreset() {
	return preset;
}

void Sanctuary::prevPreset() {
	if(preset-- == 0) preset = presets.size() - 1;
	setPreset(preset);
}

void Sanctuary::nextPreset() {
	if(++preset == presets.size()) preset = 0;
	setPreset(preset);
}

/*
 */
void Sanctuary::runBenchmark() {
	sun.reset();
	setPreset(0);
	camera.setMode(0);
	camera.setPath(0);
	setFireball(0);
	benchmark.run();
}

void Sanctuary::stopBenchmark() {
	benchmark.stop();
}

int Sanctuary::getBenchmark() {
	return benchmark.getActivity();
}

/*
 */
void Sanctuary::runPhoronix() {
	sun.reset();
	setPreset(0);
	camera.setMode(0);
	camera.setPath(0);
	setFireball(0);
	sanctuary_vbox.setHidden(1);
	phoronix.run();
}

int Sanctuary::getPhoronix() {
	return phoronix.getActivity();
}

/*
 */
void Sanctuary::playSound() {
	sound.play();
}

void Sanctuary::stopSound() {
	sound.stop();
}

/*
 */
void Sanctuary::update() {
	
	camera.update();
	
	interface.update();
	
	// console is hidden and benchmark is not running
	if(engine.console.getActivity() == 0 && benchmark.getActivity() == 0 && phoronix.getActivity() == 0) {
		
		// demo is running
		if(engine.editor.isLoaded() == 0) {
			if(camera.getMode() == 0) {
				if(gui.getKeyActivity(KEY_PAUSE) == 0 && engine.app.clearKeyState(KEY_PAUSE)) camera.setPause(!camera.getPause());
				if(gui.getKeyActivity(KEY_NEXT_PATH) == 0 && engine.app.clearKeyState(KEY_NEXT_PATH)) camera.nextPath();
			}
			if(gui.getKeyActivity(KEY_FREE_CAMERA) == 0 && engine.app.clearKeyState(KEY_FREE_CAMERA)) {
				camera.setMode(!camera.getMode());
			}
			if(gui.getKeyActivity(KEY_BENCHMARK) == 0 && engine.app.clearKeyState(KEY_BENCHMARK)) {
				benchmark_b.setFocus();
				benchmark_b.setToggled(1);
			}
		}
		
		if(gui.getKeyActivity(KEY_PREV_PRESET) == 0 && engine.app.clearKeyState(KEY_PREV_PRESET)) prevPreset();
		if(gui.getKeyActivity(KEY_NEXT_PRESET) == 0 && engine.app.clearKeyState(KEY_NEXT_PRESET)) nextPreset();
	}
	
	// flags
	int benchmark_path;
	int phoronix_path;
	int editor;
	
	// benchmark
	if(benchmark.getActivity()) {
		benchmark.update(format("%d of %d",camera.getPath() + 1,Camera::NUM_PATH));
		if(camera.getPath() == 0 && benchmark_path == Camera::NUM_PATH - 1) {
			engine.app.setMouseGrab(0);
			benchmark.done();
			#ifdef BENCHMARK
				engine.console.run("quit");
			#endif
		}
		benchmark_path = camera.getPath();
	} else {
		benchmark_path = -1;
	}
	
	// phoronix benchmark
	if(phoronix.getActivity()) {
		phoronix.update();
		if(camera.getPath() == 0 && phoronix_path == Camera::NUM_PATH - 1) {
			phoronix.done();
			#ifdef PHORONIX
				engine.console.run("quit");
			#endif
		}
		phoronix_path = camera.getPath();
	} else {
		phoronix_path = -1;
	}
	
	// demo is running
	if(engine.editor.isLoaded() == 0) {
		
		// save parameters
		if(editor == 1) {
			editor = 0;
			foreach(Flame flame; flames) flame.save();
			foreach(Candle candle; candles) candle.save();
			foreach(Fireball fireball; fireballs) fireball.save();
			sun.save();
		}
		
		// update parameters
		foreach(Flame flame; flames) flame.update();
		foreach(Candle candle; candles) candle.update();
		foreach(Fireball fireball; fireballs) fireball.update();
		sun.update();
		
		// toggle weapon
		if(camera.getMode() == 2) {
			if(getWeapon() == 0) setWeapon(1);
			weapon.update();
		} else {
			if(getWeapon()) setWeapon(0);
		}
	}
	
	// editor is loaded
	else {
		
		// restore parameters
		if(editor == 0) {
			editor = 1;
			foreach(Flame flame; flames) flame.restore();
			foreach(Candle candle; candles) candle.restore();
			foreach(Fireball fireball; fireballs) fireball.restore();
			sun.restore();
		}
		
		// stop benchmark
		if(benchmark.getActivity()) {
			benchmark.stop();
		}
	}
}

/******************************************************************************\
*
* Demo
*
\******************************************************************************/

/*
 */
Sanctuary sanctuary;

/*
 */
int init() {
	
	engine.console.run("render_hdr 0 && render_srgb 0 && render_restart");
	
	sanctuary = new Sanctuary();
	
	#ifdef BENCHMARK
		sanctuary.runBenchmark();
	#endif
	
	#ifdef PHORONIX
		sanctuary.runPhoronix();
	#endif
	
	return 1;
}

/*
 */
int shutdown() {
	
	delete sanctuary;
	
	return 1;
}

/*
 */
int update() {
	
	sanctuary.update();
	
	#ifdef _WIN32
		if(engine.app.getKeyState(APP_KEY_ALT) && engine.app.clearKeyState(APP_KEY_F4)) {
			engine.console.run("quit");
		}
	#endif
	
	if(sanctuary.getPhoronix() && engine.system.call("systemGetActivity") == 0) {
		engine.gui.setMouseCursor(GUI_CURSOR_NONE);
	}
	
	return 1;
}
