/* Copyright (C) 2005-2010, Unigine Corp. All rights reserved.
 *
 * File:    unigine.cpp
 * Desc:    Unigine system
 * Version: 1.13
 * Authors: Alexander Zaprjagaev <frustum@unigine.com>
 *          Valia Vaneeva <fattie@unigine.com>
 *
 * This file is part of the Unigine engine (http://unigine.com/).
 *
 * Your use and or redistribution of this software in source and / or
 * binary form, with or without modification, is subject to: (i) your
 * ongoing acceptance of and compliance with the terms and conditions of
 * the Unigine License Agreement; and (ii) your inclusion of this notice
 * in any version of this software that you use or redistribute.
 * A copy of the Unigine License Agreement is available by contacting
 * Unigine Corp. at http://unigine.com/
 */

#ifdef RELEASE
	#define SOFT_MOUSE
#endif

#include <unigine.h>
#include <core/scripts/system.h>
#include <core/scripts/stereo.h>

/******************************************************************************\
*
* init/shutdown
*
\******************************************************************************/

/*
 */
int init() {
	
	#ifdef RELEASE
		engine.app.setUpdate(1);
	#endif
	
	systemInit("sanctuary/unigine.ui","demos/sanctuary/locale/unigine");
	stereoInit(10.0,0.15,3.0);
	
	float aspect = float(engine.app.getWidth()) / engine.app.getHeight();
	engine.splash.setWorld(format("sanctuary/splashes/world_%s.png",(aspect < 1.5) ? "4x3" : "16x9"),16);
	engine.splash.setSplash(format("sanctuary/splashes/world_%s.png",(aspect < 1.5) ? "4x3" : "16x9"),16);
	
	engine.splash.setEnabled(1);
	engine.splash.renderSplash(0);
	engine.splash.setEnabled(0);

	engine.console.run("world_load sanctuary");
	
	return 1;
}

/*
 */
int shutdown() {
	
	#ifdef RELEASE && !BENCHMARK && !PHORONIX
		engine.console.setActivity(0);
		engine.gui.removeChild(systemGetWindow());
		float aspect = float(engine.app.getWidth()) / engine.app.getHeight();
		WidgetSprite sprite = new WidgetSprite(engine.getGui(),format("sanctuary/splashes/credits_%s.jpg",(aspect < 1.5) ? "4x3" : "16x9"));
		sprite.setBlendFunc(GUI_BLEND_NONE,GUI_BLEND_NONE);
		sprite.setTransform(scale(0.1,0.1,1.0));
		engine.gui.addChild(sprite,GUI_ALIGN_EXPAND);
		float begin = clock();
		engine.splash.setEnabled(1);
		while(clock() - begin < 5.0 && engine.app.clearKeyState(APP_KEY_ESC) == 0) {
			engine.splash.renderInterface();
		}
		engine.splash.setEnabled(0);
		delete sprite;
	#endif
	
	systemShutdown();
	stereoShutdown();
	
	return 1;
}

/******************************************************************************\
*
* update
*
\******************************************************************************/

/*
 */
int update() {
	
	systemUpdate();
	stereoUpdate();
	
	return 1;
}
